"""Type stubs for custom JAX tensor classes.

NOTE: This file is generated from templates/jax_tensors.pyi.

To regenerate, run the following from the tensor_annotations directory:
   tools/render_tensor_template.py \
     --template templates/jax_tensors.pyi \
     --out jax.pyi
"""

from typing import Any, TypeVar, Tuple, Sequence, Generic, overload, Union

from tensor_annotations.axes import Axis

A1 = TypeVar('A1', bound=Axis)
A2 = TypeVar('A2', bound=Axis)
A3 = TypeVar('A3', bound=Axis)
A4 = TypeVar('A4', bound=Axis)
A5 = TypeVar('A5', bound=Axis)
A6 = TypeVar('A6', bound=Axis)
A7 = TypeVar('A7', bound=Axis)
A8 = TypeVar('A8', bound=Axis)

Number = Union[int, float]


# A quick refresher on broadcasting rules:
# 1. Array[A, B] + scalar = Array[A, B].
# 2. Otherwise, start with trailing dimension of each tensor and work
#    forwards. Broadcasting is possible if, for each axis, the dimensions
#    of that axis in each tensor are either a) equal or b) one of them is 1.
# We deliberately ignore case b) for the time being since we don't support
# literal shapes yet.

class Array0:
  def __getitem__(self, index) -> Any: ...

  def __setitem__(self, index, value) -> Any: ...

  shape: Tuple[()]

  # BEGIN: Unary operators

  def __abs__(self) -> Array0: ...

  def __neg__(self) -> Array0: ...

  def __pos__(self) -> Array0: ...

  # END: Unary operators

  # BEGIN: Binary element-wise operators

  @overload
  def __add__(self, other: Array0) -> Array0: ...

  @overload
  def __add__(self, other: Array1) -> Array1: ...

  @overload
  def __add__(self, other: Array2) -> Array2: ...

  @overload
  def __add__(self, other: Array3) -> Array3: ...

  @overload
  def __add__(self, other: Array4) -> Array4: ...

  @overload
  def __sub__(self, other: Array0) -> Array0: ...

  @overload
  def __sub__(self, other: Array1) -> Array1: ...

  @overload
  def __sub__(self, other: Array2) -> Array2: ...

  @overload
  def __sub__(self, other: Array3) -> Array3: ...

  @overload
  def __sub__(self, other: Array4) -> Array4: ...

  @overload
  def __floordiv__(self, other: Array0) -> Array0: ...

  @overload
  def __floordiv__(self, other: Array1) -> Array1: ...

  @overload
  def __floordiv__(self, other: Array2) -> Array2: ...

  @overload
  def __floordiv__(self, other: Array3) -> Array3: ...

  @overload
  def __floordiv__(self, other: Array4) -> Array4: ...

  @overload
  def __truediv__(self, other: Array0) -> Array0: ...

  @overload
  def __truediv__(self, other: Array1) -> Array1: ...

  @overload
  def __truediv__(self, other: Array2) -> Array2: ...

  @overload
  def __truediv__(self, other: Array3) -> Array3: ...

  @overload
  def __truediv__(self, other: Array4) -> Array4: ...

  @overload
  def __pow__(self, other: Array0) -> Array0: ...

  @overload
  def __pow__(self, other: Array1) -> Array1: ...

  @overload
  def __pow__(self, other: Array2) -> Array2: ...

  @overload
  def __pow__(self, other: Array3) -> Array3: ...

  @overload
  def __pow__(self, other: Array4) -> Array4: ...

  @overload
  def __lt__(self, other: Array0) -> Array0: ...

  @overload
  def __lt__(self, other: Array1) -> Array1: ...

  @overload
  def __lt__(self, other: Array2) -> Array2: ...

  @overload
  def __lt__(self, other: Array3) -> Array3: ...

  @overload
  def __lt__(self, other: Array4) -> Array4: ...

  @overload
  def __le__(self, other: Array0) -> Array0: ...

  @overload
  def __le__(self, other: Array1) -> Array1: ...

  @overload
  def __le__(self, other: Array2) -> Array2: ...

  @overload
  def __le__(self, other: Array3) -> Array3: ...

  @overload
  def __le__(self, other: Array4) -> Array4: ...

  @overload
  def __ge__(self, other: Array0) -> Array0: ...

  @overload
  def __ge__(self, other: Array1) -> Array1: ...

  @overload
  def __ge__(self, other: Array2) -> Array2: ...

  @overload
  def __ge__(self, other: Array3) -> Array3: ...

  @overload
  def __ge__(self, other: Array4) -> Array4: ...

  @overload
  def __gt__(self, other: Array0) -> Array0: ...

  @overload
  def __gt__(self, other: Array1) -> Array1: ...

  @overload
  def __gt__(self, other: Array2) -> Array2: ...

  @overload
  def __gt__(self, other: Array3) -> Array3: ...

  @overload
  def __gt__(self, other: Array4) -> Array4: ...

  @overload
  def __mul__(self, other: Array0) -> Array0: ...

  @overload
  def __mul__(self, other: Array1) -> Array1: ...

  @overload
  def __mul__(self, other: Array2) -> Array2: ...

  @overload
  def __mul__(self, other: Array3) -> Array3: ...

  @overload
  def __mul__(self, other: Array4) -> Array4: ...

  @overload
  def __rmul__(self, other: Array0) -> Array0: ...

  @overload
  def __rmul__(self, other: Array1) -> Array1: ...

  @overload
  def __rmul__(self, other: Array2) -> Array2: ...

  @overload
  def __rmul__(self, other: Array3) -> Array3: ...

  @overload
  def __rmul__(self, other: Array4) -> Array4: ...

  # END: Binary element-wise operators


class Array1(Generic[A1]):
  def __getitem__(self, index) -> Any: ...

  def __setitem__(self, index, value) -> Any: ...

  shape: Tuple[int]

  # BEGIN: Unary operators

  def __abs__(self) -> Array1[A1]: ...

  def __neg__(self) -> Array1[A1]: ...

  def __pos__(self) -> Array1[A1]: ...

  # END: Unary operators

  # BEGIN: Binary element-wise operators

  @overload
  def __add__(self, other: Number) -> Array1[A1]: ...

  @overload
  def __add__(self, other: Array0) -> Array1[A1]: ...

  @overload
  def __add__(self, other: Array1[A1]) -> Array1[A1]: ...

  @overload
  def __sub__(self, other: Number) -> Array1[A1]: ...

  @overload
  def __sub__(self, other: Array0) -> Array1[A1]: ...

  @overload
  def __sub__(self, other: Array1[A1]) -> Array1[A1]: ...

  @overload
  def __floordiv__(self, other: Number) -> Array1[A1]: ...

  @overload
  def __floordiv__(self, other: Array0) -> Array1[A1]: ...

  @overload
  def __floordiv__(self, other: Array1[A1]) -> Array1[A1]: ...

  @overload
  def __truediv__(self, other: Number) -> Array1[A1]: ...

  @overload
  def __truediv__(self, other: Array0) -> Array1[A1]: ...

  @overload
  def __truediv__(self, other: Array1[A1]) -> Array1[A1]: ...

  @overload
  def __pow__(self, other: Number) -> Array1[A1]: ...

  @overload
  def __pow__(self, other: Array0) -> Array1[A1]: ...

  @overload
  def __pow__(self, other: Array1[A1]) -> Array1[A1]: ...

  @overload
  def __lt__(self, other: Number) -> Array1[A1]: ...

  @overload
  def __lt__(self, other: Array0) -> Array1[A1]: ...

  @overload
  def __lt__(self, other: Array1[A1]) -> Array1[A1]: ...

  @overload
  def __le__(self, other: Number) -> Array1[A1]: ...

  @overload
  def __le__(self, other: Array0) -> Array1[A1]: ...

  @overload
  def __le__(self, other: Array1[A1]) -> Array1[A1]: ...

  @overload
  def __ge__(self, other: Number) -> Array1[A1]: ...

  @overload
  def __ge__(self, other: Array0) -> Array1[A1]: ...

  @overload
  def __ge__(self, other: Array1[A1]) -> Array1[A1]: ...

  @overload
  def __gt__(self, other: Number) -> Array1[A1]: ...

  @overload
  def __gt__(self, other: Array0) -> Array1[A1]: ...

  @overload
  def __gt__(self, other: Array1[A1]) -> Array1[A1]: ...

  @overload
  def __mul__(self, other: Number) -> Array1[A1]: ...

  @overload
  def __mul__(self, other: Array0) -> Array1[A1]: ...

  @overload
  def __mul__(self, other: Array1[A1]) -> Array1[A1]: ...

  @overload
  def __rmul__(self, other: Number) -> Array1[A1]: ...

  @overload
  def __rmul__(self, other: Array0) -> Array1[A1]: ...

  @overload
  def __rmul__(self, other: Array1[A1]) -> Array1[A1]: ...

  # END: Binary element-wise operators


class Array2(Generic[A1, A2]):
  def __getitem__(self, index) -> Any: ...

  def __setitem__(self, index, value) -> Any: ...

  shape: Tuple[int, int]

  # BEGIN: Unary operators

  def __abs__(self) -> Array2[A1, A2]: ...

  def __neg__(self) -> Array2[A1, A2]: ...

  def __pos__(self) -> Array2[A1, A2]: ...

  # END: Unary operators

  # BEGIN: Binary element-wise operators

  @overload
  def __add__(self, other: Number) -> Array2[A1, A2]: ...

  @overload
  def __add__(self, other: Array0) -> Array2[A1, A2]: ...

  @overload
  def __add__(self, other: Array1[A2]) -> Array2[A1, A2]: ...

  @overload
  def __add__(self, other: Array2[A1, A2]) -> Array2[A1, A2]: ...

  @overload
  def __sub__(self, other: Number) -> Array2[A1, A2]: ...

  @overload
  def __sub__(self, other: Array0) -> Array2[A1, A2]: ...

  @overload
  def __sub__(self, other: Array1[A2]) -> Array2[A1, A2]: ...

  @overload
  def __sub__(self, other: Array2[A1, A2]) -> Array2[A1, A2]: ...

  @overload
  def __floordiv__(self, other: Number) -> Array2[A1, A2]: ...

  @overload
  def __floordiv__(self, other: Array0) -> Array2[A1, A2]: ...

  @overload
  def __floordiv__(self, other: Array1[A2]) -> Array2[A1, A2]: ...

  @overload
  def __floordiv__(self, other: Array2[A1, A2]) -> Array2[A1, A2]: ...

  @overload
  def __truediv__(self, other: Number) -> Array2[A1, A2]: ...

  @overload
  def __truediv__(self, other: Array0) -> Array2[A1, A2]: ...

  @overload
  def __truediv__(self, other: Array1[A2]) -> Array2[A1, A2]: ...

  @overload
  def __truediv__(self, other: Array2[A1, A2]) -> Array2[A1, A2]: ...

  @overload
  def __pow__(self, other: Number) -> Array2[A1, A2]: ...

  @overload
  def __pow__(self, other: Array0) -> Array2[A1, A2]: ...

  @overload
  def __pow__(self, other: Array1[A2]) -> Array2[A1, A2]: ...

  @overload
  def __pow__(self, other: Array2[A1, A2]) -> Array2[A1, A2]: ...

  @overload
  def __lt__(self, other: Number) -> Array2[A1, A2]: ...

  @overload
  def __lt__(self, other: Array0) -> Array2[A1, A2]: ...

  @overload
  def __lt__(self, other: Array1[A2]) -> Array2[A1, A2]: ...

  @overload
  def __lt__(self, other: Array2[A1, A2]) -> Array2[A1, A2]: ...

  @overload
  def __le__(self, other: Number) -> Array2[A1, A2]: ...

  @overload
  def __le__(self, other: Array0) -> Array2[A1, A2]: ...

  @overload
  def __le__(self, other: Array1[A2]) -> Array2[A1, A2]: ...

  @overload
  def __le__(self, other: Array2[A1, A2]) -> Array2[A1, A2]: ...

  @overload
  def __ge__(self, other: Number) -> Array2[A1, A2]: ...

  @overload
  def __ge__(self, other: Array0) -> Array2[A1, A2]: ...

  @overload
  def __ge__(self, other: Array1[A2]) -> Array2[A1, A2]: ...

  @overload
  def __ge__(self, other: Array2[A1, A2]) -> Array2[A1, A2]: ...

  @overload
  def __gt__(self, other: Number) -> Array2[A1, A2]: ...

  @overload
  def __gt__(self, other: Array0) -> Array2[A1, A2]: ...

  @overload
  def __gt__(self, other: Array1[A2]) -> Array2[A1, A2]: ...

  @overload
  def __gt__(self, other: Array2[A1, A2]) -> Array2[A1, A2]: ...

  @overload
  def __mul__(self, other: Number) -> Array2[A1, A2]: ...

  @overload
  def __mul__(self, other: Array0) -> Array2[A1, A2]: ...

  @overload
  def __mul__(self, other: Array1[A2]) -> Array2[A1, A2]: ...

  @overload
  def __mul__(self, other: Array2[A1, A2]) -> Array2[A1, A2]: ...

  @overload
  def __rmul__(self, other: Number) -> Array2[A1, A2]: ...

  @overload
  def __rmul__(self, other: Array0) -> Array2[A1, A2]: ...

  @overload
  def __rmul__(self, other: Array1[A2]) -> Array2[A1, A2]: ...

  @overload
  def __rmul__(self, other: Array2[A1, A2]) -> Array2[A1, A2]: ...

  # END: Binary element-wise operators


class Array3(Generic[A1, A2, A3]):
  def __getitem__(self, index) -> Any: ...

  def __setitem__(self, index, value) -> Any: ...

  shape: Tuple[int, int, int]

  # BEGIN: Unary operators

  def __abs__(self) -> Array3[A1, A2, A3]: ...

  def __neg__(self) -> Array3[A1, A2, A3]: ...

  def __pos__(self) -> Array3[A1, A2, A3]: ...

  # END: Unary operators

  # BEGIN: Binary element-wise operators

  @overload
  def __add__(self, other: Number) -> Array3[A1, A2, A3]: ...

  @overload
  def __add__(self, other: Array0) -> Array3[A1, A2, A3]: ...

  @overload
  def __add__(self, other: Array1[A3]) -> Array3[A1, A2, A3]: ...

  @overload
  def __add__(self, other: Array2[A2, A3]) -> Array3[A1, A2, A3]: ...

  @overload
  def __add__(self, other: Array3[A1, A2, A3]) -> Array3[A1, A2, A3]: ...

  @overload
  def __sub__(self, other: Number) -> Array3[A1, A2, A3]: ...

  @overload
  def __sub__(self, other: Array0) -> Array3[A1, A2, A3]: ...

  @overload
  def __sub__(self, other: Array1[A3]) -> Array3[A1, A2, A3]: ...

  @overload
  def __sub__(self, other: Array2[A2, A3]) -> Array3[A1, A2, A3]: ...

  @overload
  def __sub__(self, other: Array3[A1, A2, A3]) -> Array3[A1, A2, A3]: ...

  @overload
  def __floordiv__(self, other: Number) -> Array3[A1, A2, A3]: ...

  @overload
  def __floordiv__(self, other: Array0) -> Array3[A1, A2, A3]: ...

  @overload
  def __floordiv__(self, other: Array1[A3]) -> Array3[A1, A2, A3]: ...

  @overload
  def __floordiv__(self, other: Array2[A2, A3]) -> Array3[A1, A2, A3]: ...

  @overload
  def __floordiv__(self, other: Array3[A1, A2, A3]) -> Array3[A1, A2, A3]: ...

  @overload
  def __truediv__(self, other: Number) -> Array3[A1, A2, A3]: ...

  @overload
  def __truediv__(self, other: Array0) -> Array3[A1, A2, A3]: ...

  @overload
  def __truediv__(self, other: Array1[A3]) -> Array3[A1, A2, A3]: ...

  @overload
  def __truediv__(self, other: Array2[A2, A3]) -> Array3[A1, A2, A3]: ...

  @overload
  def __truediv__(self, other: Array3[A1, A2, A3]) -> Array3[A1, A2, A3]: ...

  @overload
  def __pow__(self, other: Number) -> Array3[A1, A2, A3]: ...

  @overload
  def __pow__(self, other: Array0) -> Array3[A1, A2, A3]: ...

  @overload
  def __pow__(self, other: Array1[A3]) -> Array3[A1, A2, A3]: ...

  @overload
  def __pow__(self, other: Array2[A2, A3]) -> Array3[A1, A2, A3]: ...

  @overload
  def __pow__(self, other: Array3[A1, A2, A3]) -> Array3[A1, A2, A3]: ...

  @overload
  def __lt__(self, other: Number) -> Array3[A1, A2, A3]: ...

  @overload
  def __lt__(self, other: Array0) -> Array3[A1, A2, A3]: ...

  @overload
  def __lt__(self, other: Array1[A3]) -> Array3[A1, A2, A3]: ...

  @overload
  def __lt__(self, other: Array2[A2, A3]) -> Array3[A1, A2, A3]: ...

  @overload
  def __lt__(self, other: Array3[A1, A2, A3]) -> Array3[A1, A2, A3]: ...

  @overload
  def __le__(self, other: Number) -> Array3[A1, A2, A3]: ...

  @overload
  def __le__(self, other: Array0) -> Array3[A1, A2, A3]: ...

  @overload
  def __le__(self, other: Array1[A3]) -> Array3[A1, A2, A3]: ...

  @overload
  def __le__(self, other: Array2[A2, A3]) -> Array3[A1, A2, A3]: ...

  @overload
  def __le__(self, other: Array3[A1, A2, A3]) -> Array3[A1, A2, A3]: ...

  @overload
  def __ge__(self, other: Number) -> Array3[A1, A2, A3]: ...

  @overload
  def __ge__(self, other: Array0) -> Array3[A1, A2, A3]: ...

  @overload
  def __ge__(self, other: Array1[A3]) -> Array3[A1, A2, A3]: ...

  @overload
  def __ge__(self, other: Array2[A2, A3]) -> Array3[A1, A2, A3]: ...

  @overload
  def __ge__(self, other: Array3[A1, A2, A3]) -> Array3[A1, A2, A3]: ...

  @overload
  def __gt__(self, other: Number) -> Array3[A1, A2, A3]: ...

  @overload
  def __gt__(self, other: Array0) -> Array3[A1, A2, A3]: ...

  @overload
  def __gt__(self, other: Array1[A3]) -> Array3[A1, A2, A3]: ...

  @overload
  def __gt__(self, other: Array2[A2, A3]) -> Array3[A1, A2, A3]: ...

  @overload
  def __gt__(self, other: Array3[A1, A2, A3]) -> Array3[A1, A2, A3]: ...

  @overload
  def __mul__(self, other: Number) -> Array3[A1, A2, A3]: ...

  @overload
  def __mul__(self, other: Array0) -> Array3[A1, A2, A3]: ...

  @overload
  def __mul__(self, other: Array1[A3]) -> Array3[A1, A2, A3]: ...

  @overload
  def __mul__(self, other: Array2[A2, A3]) -> Array3[A1, A2, A3]: ...

  @overload
  def __mul__(self, other: Array3[A1, A2, A3]) -> Array3[A1, A2, A3]: ...

  @overload
  def __rmul__(self, other: Number) -> Array3[A1, A2, A3]: ...

  @overload
  def __rmul__(self, other: Array0) -> Array3[A1, A2, A3]: ...

  @overload
  def __rmul__(self, other: Array1[A3]) -> Array3[A1, A2, A3]: ...

  @overload
  def __rmul__(self, other: Array2[A2, A3]) -> Array3[A1, A2, A3]: ...

  @overload
  def __rmul__(self, other: Array3[A1, A2, A3]) -> Array3[A1, A2, A3]: ...

  # END: Binary element-wise operators


class Array4(Generic[A1, A2, A3, A4]):
  def __getitem__(self, index) -> Any: ...

  def __setitem__(self, index, value) -> Any: ...

  shape: Tuple[int, int, int, int]

  # BEGIN: Unary operators

  def __abs__(self) -> Array4[A1, A2, A3, A4]: ...

  def __neg__(self) -> Array4[A1, A2, A3, A4]: ...

  def __pos__(self) -> Array4[A1, A2, A3, A4]: ...

  # END: Unary operators

  # BEGIN: Binary element-wise operators

  @overload
  def __add__(self, other: Number) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __add__(self, other: Array0) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __add__(self, other: Array1[A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __add__(self, other: Array2[A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __add__(self, other: Array3[A2, A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __add__(self, other: Array4[A1, A2, A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __sub__(self, other: Number) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __sub__(self, other: Array0) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __sub__(self, other: Array1[A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __sub__(self, other: Array2[A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __sub__(self, other: Array3[A2, A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __sub__(self, other: Array4[A1, A2, A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __floordiv__(self, other: Number) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __floordiv__(self, other: Array0) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __floordiv__(self, other: Array1[A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __floordiv__(self, other: Array2[A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __floordiv__(self, other: Array3[A2, A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __floordiv__(self, other: Array4[A1, A2, A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __truediv__(self, other: Number) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __truediv__(self, other: Array0) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __truediv__(self, other: Array1[A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __truediv__(self, other: Array2[A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __truediv__(self, other: Array3[A2, A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __truediv__(self, other: Array4[A1, A2, A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __pow__(self, other: Number) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __pow__(self, other: Array0) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __pow__(self, other: Array1[A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __pow__(self, other: Array2[A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __pow__(self, other: Array3[A2, A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __pow__(self, other: Array4[A1, A2, A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __lt__(self, other: Number) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __lt__(self, other: Array0) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __lt__(self, other: Array1[A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __lt__(self, other: Array2[A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __lt__(self, other: Array3[A2, A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __lt__(self, other: Array4[A1, A2, A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __le__(self, other: Number) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __le__(self, other: Array0) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __le__(self, other: Array1[A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __le__(self, other: Array2[A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __le__(self, other: Array3[A2, A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __le__(self, other: Array4[A1, A2, A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __ge__(self, other: Number) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __ge__(self, other: Array0) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __ge__(self, other: Array1[A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __ge__(self, other: Array2[A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __ge__(self, other: Array3[A2, A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __ge__(self, other: Array4[A1, A2, A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __gt__(self, other: Number) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __gt__(self, other: Array0) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __gt__(self, other: Array1[A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __gt__(self, other: Array2[A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __gt__(self, other: Array3[A2, A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __gt__(self, other: Array4[A1, A2, A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __mul__(self, other: Number) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __mul__(self, other: Array0) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __mul__(self, other: Array1[A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __mul__(self, other: Array2[A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __mul__(self, other: Array3[A2, A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __mul__(self, other: Array4[A1, A2, A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __rmul__(self, other: Number) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __rmul__(self, other: Array0) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __rmul__(self, other: Array1[A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __rmul__(self, other: Array2[A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __rmul__(self, other: Array3[A2, A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  @overload
  def __rmul__(self, other: Array4[A1, A2, A3, A4]) -> Array4[A1, A2, A3, A4]: ...

  # END: Binary element-wise operators


class Array5(Generic[A1, A2, A3, A4, A5]):
  def __getitem__(self, index) -> Any: ...

  def __setitem__(self, index, value) -> Any: ...

  shape: Tuple[int, int, int, int]

  # BEGIN: Unary operators

  def __abs__(self) -> Array5[A1, A2, A3, A4, A5]: ...

  def __neg__(self) -> Array5[A1, A2, A3, A4, A5]: ...

  def __pos__(self) -> Array5[A1, A2, A3, A4, A5]: ...

  # END: Unary operators

  # BEGIN: Binary element-wise operators

  @overload
  def __add__(self, other: Number) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __add__(self, other: Array0) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __add__(self, other: Array1[A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __add__(self, other: Array2[A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __add__(self, other: Array3[A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __add__(self, other: Array4[A2, A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __add__(self, other: Array5[A1, A2, A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __sub__(self, other: Number) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __sub__(self, other: Array0) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __sub__(self, other: Array1[A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __sub__(self, other: Array2[A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __sub__(self, other: Array3[A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __sub__(self, other: Array4[A2, A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __sub__(self, other: Array5[A1, A2, A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __floordiv__(self, other: Number) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __floordiv__(self, other: Array0) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __floordiv__(self, other: Array1[A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __floordiv__(self, other: Array2[A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __floordiv__(self, other: Array3[A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __floordiv__(self, other: Array4[A2, A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __floordiv__(self, other: Array5[A1, A2, A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __truediv__(self, other: Number) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __truediv__(self, other: Array0) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __truediv__(self, other: Array1[A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __truediv__(self, other: Array2[A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __truediv__(self, other: Array3[A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __truediv__(self, other: Array4[A2, A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __truediv__(self, other: Array5[A1, A2, A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __pow__(self, other: Number) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __pow__(self, other: Array0) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __pow__(self, other: Array1[A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __pow__(self, other: Array2[A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __pow__(self, other: Array3[A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __pow__(self, other: Array4[A2, A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __pow__(self, other: Array5[A1, A2, A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __lt__(self, other: Number) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __lt__(self, other: Array0) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __lt__(self, other: Array1[A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __lt__(self, other: Array2[A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __lt__(self, other: Array3[A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __lt__(self, other: Array4[A2, A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __lt__(self, other: Array5[A1, A2, A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __le__(self, other: Number) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __le__(self, other: Array0) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __le__(self, other: Array1[A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __le__(self, other: Array2[A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __le__(self, other: Array3[A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __le__(self, other: Array4[A2, A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __le__(self, other: Array5[A1, A2, A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __ge__(self, other: Number) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __ge__(self, other: Array0) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __ge__(self, other: Array1[A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __ge__(self, other: Array2[A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __ge__(self, other: Array3[A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __ge__(self, other: Array4[A2, A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __ge__(self, other: Array5[A1, A2, A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __gt__(self, other: Number) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __gt__(self, other: Array0) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __gt__(self, other: Array1[A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __gt__(self, other: Array2[A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __gt__(self, other: Array3[A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __gt__(self, other: Array4[A2, A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __gt__(self, other: Array5[A1, A2, A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __mul__(self, other: Number) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __mul__(self, other: Array0) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __mul__(self, other: Array1[A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __mul__(self, other: Array2[A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __mul__(self, other: Array3[A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __mul__(self, other: Array4[A2, A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __mul__(self, other: Array5[A1, A2, A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __rmul__(self, other: Number) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __rmul__(self, other: Array0) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __rmul__(self, other: Array1[A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __rmul__(self, other: Array2[A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __rmul__(self, other: Array3[A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __rmul__(self, other: Array4[A2, A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  @overload
  def __rmul__(self, other: Array5[A1, A2, A3, A4, A5]) -> Array5[A1, A2, A3, A4, A5]: ...

  # END: Binary element-wise operators


class Array6(Generic[A1, A2, A3, A4, A5, A6]):
  def __getitem__(self, index) -> Any: ...

  def __setitem__(self, index, value) -> Any: ...

  shape: Tuple[int, int, int, int]

  # BEGIN: Unary operators

  def __abs__(self) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  def __neg__(self) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  def __pos__(self) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  # END: Unary operators

  # BEGIN: Binary element-wise operators

  @overload
  def __add__(self, other: Number) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __add__(self, other: Array0) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __add__(self, other: Array1[A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __add__(self, other: Array2[A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __add__(self, other: Array3[A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __add__(self, other: Array4[A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __add__(self, other: Array5[A2, A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __add__(self, other: Array6[A1, A2, A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __sub__(self, other: Number) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __sub__(self, other: Array0) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __sub__(self, other: Array1[A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __sub__(self, other: Array2[A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __sub__(self, other: Array3[A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __sub__(self, other: Array4[A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __sub__(self, other: Array5[A2, A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __sub__(self, other: Array6[A1, A2, A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __floordiv__(self, other: Number) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __floordiv__(self, other: Array0) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __floordiv__(self, other: Array1[A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __floordiv__(self, other: Array2[A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __floordiv__(self, other: Array3[A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __floordiv__(self, other: Array4[A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __floordiv__(self, other: Array5[A2, A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __floordiv__(self, other: Array6[A1, A2, A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __truediv__(self, other: Number) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __truediv__(self, other: Array0) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __truediv__(self, other: Array1[A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __truediv__(self, other: Array2[A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __truediv__(self, other: Array3[A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __truediv__(self, other: Array4[A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __truediv__(self, other: Array5[A2, A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __truediv__(self, other: Array6[A1, A2, A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __pow__(self, other: Number) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __pow__(self, other: Array0) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __pow__(self, other: Array1[A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __pow__(self, other: Array2[A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __pow__(self, other: Array3[A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __pow__(self, other: Array4[A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __pow__(self, other: Array5[A2, A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __pow__(self, other: Array6[A1, A2, A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __lt__(self, other: Number) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __lt__(self, other: Array0) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __lt__(self, other: Array1[A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __lt__(self, other: Array2[A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __lt__(self, other: Array3[A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __lt__(self, other: Array4[A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __lt__(self, other: Array5[A2, A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __lt__(self, other: Array6[A1, A2, A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __le__(self, other: Number) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __le__(self, other: Array0) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __le__(self, other: Array1[A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __le__(self, other: Array2[A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __le__(self, other: Array3[A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __le__(self, other: Array4[A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __le__(self, other: Array5[A2, A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __le__(self, other: Array6[A1, A2, A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __ge__(self, other: Number) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __ge__(self, other: Array0) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __ge__(self, other: Array1[A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __ge__(self, other: Array2[A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __ge__(self, other: Array3[A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __ge__(self, other: Array4[A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __ge__(self, other: Array5[A2, A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __ge__(self, other: Array6[A1, A2, A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __gt__(self, other: Number) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __gt__(self, other: Array0) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __gt__(self, other: Array1[A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __gt__(self, other: Array2[A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __gt__(self, other: Array3[A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __gt__(self, other: Array4[A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __gt__(self, other: Array5[A2, A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __gt__(self, other: Array6[A1, A2, A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __mul__(self, other: Number) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __mul__(self, other: Array0) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __mul__(self, other: Array1[A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __mul__(self, other: Array2[A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __mul__(self, other: Array3[A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __mul__(self, other: Array4[A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __mul__(self, other: Array5[A2, A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __mul__(self, other: Array6[A1, A2, A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __rmul__(self, other: Number) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __rmul__(self, other: Array0) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __rmul__(self, other: Array1[A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __rmul__(self, other: Array2[A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __rmul__(self, other: Array3[A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __rmul__(self, other: Array4[A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __rmul__(self, other: Array5[A2, A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  @overload
  def __rmul__(self, other: Array6[A1, A2, A3, A4, A5, A6]) -> Array6[A1, A2, A3, A4, A5, A6]: ...

  # END: Binary element-wise operators


class Array7(Generic[A1, A2, A3, A4, A5, A6, A7]):
  def __getitem__(self, index) -> Any: ...

  def __setitem__(self, index, value) -> Any: ...

  shape: Tuple[int, int, int, int]

  # BEGIN: Unary operators

  def __abs__(self) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  def __neg__(self) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  def __pos__(self) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  # END: Unary operators

  # BEGIN: Binary element-wise operators

  @overload
  def __add__(self, other: Number) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __add__(self, other: Array0) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __add__(self, other: Array1[A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __add__(self, other: Array2[A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __add__(self, other: Array3[A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __add__(self, other: Array4[A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __add__(self, other: Array5[A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __add__(self, other: Array6[A2, A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __add__(self, other: Array7[A1, A2, A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __sub__(self, other: Number) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __sub__(self, other: Array0) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __sub__(self, other: Array1[A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __sub__(self, other: Array2[A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __sub__(self, other: Array3[A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __sub__(self, other: Array4[A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __sub__(self, other: Array5[A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __sub__(self, other: Array6[A2, A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __sub__(self, other: Array7[A1, A2, A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __floordiv__(self, other: Number) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __floordiv__(self, other: Array0) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __floordiv__(self, other: Array1[A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __floordiv__(self, other: Array2[A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __floordiv__(self, other: Array3[A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __floordiv__(self, other: Array4[A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __floordiv__(self, other: Array5[A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __floordiv__(self, other: Array6[A2, A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __floordiv__(self, other: Array7[A1, A2, A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __truediv__(self, other: Number) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __truediv__(self, other: Array0) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __truediv__(self, other: Array1[A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __truediv__(self, other: Array2[A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __truediv__(self, other: Array3[A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __truediv__(self, other: Array4[A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __truediv__(self, other: Array5[A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __truediv__(self, other: Array6[A2, A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __truediv__(self, other: Array7[A1, A2, A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __pow__(self, other: Number) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __pow__(self, other: Array0) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __pow__(self, other: Array1[A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __pow__(self, other: Array2[A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __pow__(self, other: Array3[A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __pow__(self, other: Array4[A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __pow__(self, other: Array5[A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __pow__(self, other: Array6[A2, A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __pow__(self, other: Array7[A1, A2, A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __lt__(self, other: Number) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __lt__(self, other: Array0) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __lt__(self, other: Array1[A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __lt__(self, other: Array2[A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __lt__(self, other: Array3[A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __lt__(self, other: Array4[A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __lt__(self, other: Array5[A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __lt__(self, other: Array6[A2, A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __lt__(self, other: Array7[A1, A2, A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __le__(self, other: Number) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __le__(self, other: Array0) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __le__(self, other: Array1[A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __le__(self, other: Array2[A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __le__(self, other: Array3[A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __le__(self, other: Array4[A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __le__(self, other: Array5[A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __le__(self, other: Array6[A2, A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __le__(self, other: Array7[A1, A2, A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __ge__(self, other: Number) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __ge__(self, other: Array0) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __ge__(self, other: Array1[A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __ge__(self, other: Array2[A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __ge__(self, other: Array3[A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __ge__(self, other: Array4[A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __ge__(self, other: Array5[A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __ge__(self, other: Array6[A2, A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __ge__(self, other: Array7[A1, A2, A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __gt__(self, other: Number) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __gt__(self, other: Array0) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __gt__(self, other: Array1[A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __gt__(self, other: Array2[A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __gt__(self, other: Array3[A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __gt__(self, other: Array4[A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __gt__(self, other: Array5[A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __gt__(self, other: Array6[A2, A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __gt__(self, other: Array7[A1, A2, A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __mul__(self, other: Number) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __mul__(self, other: Array0) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __mul__(self, other: Array1[A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __mul__(self, other: Array2[A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __mul__(self, other: Array3[A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __mul__(self, other: Array4[A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __mul__(self, other: Array5[A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __mul__(self, other: Array6[A2, A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __mul__(self, other: Array7[A1, A2, A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __rmul__(self, other: Number) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __rmul__(self, other: Array0) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __rmul__(self, other: Array1[A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __rmul__(self, other: Array2[A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __rmul__(self, other: Array3[A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __rmul__(self, other: Array4[A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __rmul__(self, other: Array5[A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __rmul__(self, other: Array6[A2, A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  @overload
  def __rmul__(self, other: Array7[A1, A2, A3, A4, A5, A6, A7]) -> Array7[A1, A2, A3, A4, A5, A6, A7]: ...

  # END: Binary element-wise operators


class Array8(Generic[A1, A2, A3, A4, A5, A6, A7, A8]):
  def __getitem__(self, index) -> Any: ...

  def __setitem__(self, index, value) -> Any: ...

  shape: Tuple[int, int, int, int]

  # BEGIN: Unary operators

  def __abs__(self) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  def __neg__(self) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  def __pos__(self) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  # END: Unary operators

  # BEGIN: Binary element-wise operators

  @overload
  def __add__(self, other: Number) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __add__(self, other: Array0) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __add__(self, other: Array1[A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __add__(self, other: Array2[A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __add__(self, other: Array3[A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __add__(self, other: Array4[A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __add__(self, other: Array5[A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __add__(self, other: Array6[A3, A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __add__(self, other: Array7[A2, A3, A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __add__(self, other: Array8[A1, A2, A3, A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __sub__(self, other: Number) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __sub__(self, other: Array0) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __sub__(self, other: Array1[A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __sub__(self, other: Array2[A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __sub__(self, other: Array3[A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __sub__(self, other: Array4[A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __sub__(self, other: Array5[A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __sub__(self, other: Array6[A3, A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __sub__(self, other: Array7[A2, A3, A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __sub__(self, other: Array8[A1, A2, A3, A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __floordiv__(self, other: Number) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __floordiv__(self, other: Array0) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __floordiv__(self, other: Array1[A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __floordiv__(self, other: Array2[A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __floordiv__(self, other: Array3[A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __floordiv__(self, other: Array4[A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __floordiv__(self, other: Array5[A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __floordiv__(self, other: Array6[A3, A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __floordiv__(self, other: Array7[A2, A3, A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __floordiv__(self, other: Array8[A1, A2, A3, A4, A5, A6, A7, A8]) -> \
  Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __truediv__(self, other: Number) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __truediv__(self, other: Array0) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __truediv__(self, other: Array1[A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __truediv__(self, other: Array2[A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __truediv__(self, other: Array3[A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __truediv__(self, other: Array4[A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __truediv__(self, other: Array5[A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __truediv__(self, other: Array6[A3, A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __truediv__(self, other: Array7[A2, A3, A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __truediv__(self, other: Array8[A1, A2, A3, A4, A5, A6, A7, A8]) -> \
  Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __pow__(self, other: Number) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __pow__(self, other: Array0) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __pow__(self, other: Array1[A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __pow__(self, other: Array2[A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __pow__(self, other: Array3[A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __pow__(self, other: Array4[A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __pow__(self, other: Array5[A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __pow__(self, other: Array6[A3, A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __pow__(self, other: Array7[A2, A3, A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __pow__(self, other: Array8[A1, A2, A3, A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __lt__(self, other: Number) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __lt__(self, other: Array0) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __lt__(self, other: Array1[A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __lt__(self, other: Array2[A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __lt__(self, other: Array3[A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __lt__(self, other: Array4[A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __lt__(self, other: Array5[A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __lt__(self, other: Array6[A3, A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __lt__(self, other: Array7[A2, A3, A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __lt__(self, other: Array8[A1, A2, A3, A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __le__(self, other: Number) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __le__(self, other: Array0) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __le__(self, other: Array1[A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __le__(self, other: Array2[A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __le__(self, other: Array3[A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __le__(self, other: Array4[A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __le__(self, other: Array5[A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __le__(self, other: Array6[A3, A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __le__(self, other: Array7[A2, A3, A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __le__(self, other: Array8[A1, A2, A3, A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __ge__(self, other: Number) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __ge__(self, other: Array0) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __ge__(self, other: Array1[A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __ge__(self, other: Array2[A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __ge__(self, other: Array3[A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __ge__(self, other: Array4[A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __ge__(self, other: Array5[A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __ge__(self, other: Array6[A3, A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __ge__(self, other: Array7[A2, A3, A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __ge__(self, other: Array8[A1, A2, A3, A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __gt__(self, other: Number) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __gt__(self, other: Array0) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __gt__(self, other: Array1[A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __gt__(self, other: Array2[A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __gt__(self, other: Array3[A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __gt__(self, other: Array4[A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __gt__(self, other: Array5[A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __gt__(self, other: Array6[A3, A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __gt__(self, other: Array7[A2, A3, A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __gt__(self, other: Array8[A1, A2, A3, A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __mul__(self, other: Number) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __mul__(self, other: Array0) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __mul__(self, other: Array1[A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __mul__(self, other: Array2[A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __mul__(self, other: Array3[A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __mul__(self, other: Array4[A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __mul__(self, other: Array5[A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __mul__(self, other: Array6[A3, A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __mul__(self, other: Array7[A2, A3, A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __mul__(self, other: Array8[A1, A2, A3, A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __rmul__(self, other: Number) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __rmul__(self, other: Array0) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __rmul__(self, other: Array1[A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __rmul__(self, other: Array2[A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __rmul__(self, other: Array3[A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __rmul__(self, other: Array4[A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __rmul__(self, other: Array5[A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __rmul__(self, other: Array6[A3, A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __rmul__(self, other: Array7[A2, A3, A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  @overload
  def __rmul__(self, other: Array8[A1, A2, A3, A4, A5, A6, A7, A8]) -> Array8[A1, A2, A3, A4, A5, A6, A7, A8]: ...

  # END: Binary element-wise operators



# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Canonical source for common axis types.

We need to make sure that `Batch` in one module means the same as `Batch` in
another module. Since we verify equality of axis types based on the identity of
the type, that means that both modules need to import and use the same `Batch`
type. We therefore provide this file as the canonical reference for axis types
that are likely to be used widely.
"""


class Axis:
  """Base type for axis annotations.

  User-defined axis types should subclass this.
  """
  pass


# These could be more compactly specified with typing.NewType,
# but pytype doesn't currently support NewType in stubs:
# https://github.com/google/pytype/issues/597


class Batch(Axis):
  pass


class Channels(Axis):
  pass


class Features(Axis):
  pass


class Time(Axis):
  pass


class Height(Axis):
  pass


class Width(Axis):
  pass


class Depth(Axis):
  pass



# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for TensorFlow stubs."""

from typing import NewType
import unittest
from tensor_annotations import axes
from tensor_annotations.tensorflow import Tensor0
from tensor_annotations.tensorflow import Tensor1
from tensor_annotations.tensorflow import Tensor2
from tensor_annotations.tests import utils
import tensorflow as tf


A1 = NewType('A1', axes.Axis)
A2 = NewType('A2', axes.Axis)

# It's less than ideal that we have to repeat imports etc. here for pytype, but
# this seems like the best balance between readability and complexity.
_PREAMBLE = """
from typing import NewType

import tensorflow as tf
from tensor_annotations import axes
from tensor_annotations.tensorflow import Tensor0, Tensor1, Tensor2

A1 = NewType('A1', axes.Axis)
A2 = NewType('A2', axes.Axis)
"""


class TensorFlowStubTests(unittest.TestCase):

  def testTranspose_InferredMatchesActualShapeShape(self):
    with utils.SaveCodeAsString() as code_saver:
      x: Tensor2[A1, A2] = tf.zeros((1, 2))
      y = tf.transpose(x)

    inferred = utils.pytype_infer_shapes(_PREAMBLE + code_saver.code)

    self.assertEqual(inferred.y, y.shape)

  def testUnaryOperator_ReturnCustomType(self):
    with utils.SaveCodeAsString() as code_saver:
      x: Tensor1[A1] = tf.zeros((1,))
      # Let's just test a representative subset.
      a = tf.abs(x)  # pylint: disable=unused-variable
      b = tf.sin(x)  # pylint: disable=unused-variable
      c = tf.floor(x)  # pylint: disable=unused-variable
      d = tf.ones_like(x)  # pylint: disable=unused-variable
      e = tf.round(x)  # pylint: disable=unused-variable
      f = tf.sign(x)  # pylint: disable=unused-variable

    inferred = utils.pytype_infer_types(_PREAMBLE + code_saver.code)

    expected = 'Tensor1[A1]'
    self.assertEqual(inferred.a, expected)
    self.assertEqual(inferred.b, expected)
    self.assertEqual(inferred.c, expected)
    self.assertEqual(inferred.d, expected)
    self.assertEqual(inferred.e, expected)
    self.assertEqual(inferred.f, expected)

  def testMathUnaryOperator_ReturnCustomType(self):
    with utils.SaveCodeAsString() as code_saver:
      x: Tensor1[A1] = tf.zeros((1,))
      # Let's just test a representative subset.
      a = tf.math.abs(x)  # pylint: disable=unused-variable
      b = tf.math.sin(x)  # pylint: disable=unused-variable
      c = tf.math.floor(x)  # pylint: disable=unused-variable
      d = tf.math.round(x)  # pylint: disable=unused-variable
      e = tf.math.sign(x)  # pylint: disable=unused-variable

    inferred = utils.pytype_infer_types(_PREAMBLE + code_saver.code)

    expected = 'Tensor1[A1]'
    self.assertEqual(inferred.a, expected)
    self.assertEqual(inferred.b, expected)
    self.assertEqual(inferred.c, expected)
    self.assertEqual(inferred.d, expected)
    self.assertEqual(inferred.e, expected)

  def testZerosOnes_ReturnsCorrectShape(self):
    with utils.SaveCodeAsString() as code_saver:
      a = tf.zeros(())  # pylint: disable=unused-variable
      b = tf.ones(())  # pylint: disable=unused-variable
      c = tf.zeros((1,))  # pylint: disable=unused-variable
      d = tf.ones((1,))  # pylint: disable=unused-variable
      e = tf.zeros((1, 1))  # pylint: disable=unused-variable
      f = tf.ones((1, 1))  # pylint: disable=unused-variable

    inferred = utils.pytype_infer_types(_PREAMBLE + code_saver.code)

    self.assertEqual(inferred.a, 'Tensor0')
    self.assertEqual(inferred.b, 'Tensor0')
    self.assertEqual(inferred.c, 'Tensor1')
    self.assertEqual(inferred.d, 'Tensor1')
    self.assertEqual(inferred.e, 'Tensor2')
    self.assertEqual(inferred.f, 'Tensor2')

  def testSum_InferredMatchesActualShape(self):
    with utils.SaveCodeAsString() as code_saver:
      x: Tensor1[A1] = tf.zeros((1,))
      y: Tensor2[A1, A2] = tf.zeros((1, 2))
      x0 = tf.reduce_sum(x, axis=0)
      y0 = tf.reduce_sum(y, axis=0)
      y1 = tf.reduce_sum(y, axis=1)

    inferred = utils.pytype_infer_shapes(_PREAMBLE + code_saver.code)

    self.assertEqual(inferred.x0, x0.shape)
    self.assertEqual(inferred.y0, y0.shape)
    self.assertEqual(inferred.y1, y1.shape)

  def testTensorAdd_ReturnsCustomType(self):
    with utils.SaveCodeAsString() as code_saver:
      x: Tensor1[A1] = tf.zeros((1,))
      a = x + 1  # pylint: disable=unused-variable
      b = x + x  # pylint: disable=unused-variable

    inferred = utils.pytype_infer_types(_PREAMBLE + code_saver.code)

    self.assertEqual(inferred.a, 'Tensor1[A1]')
    self.assertEqual(inferred.b, 'Tensor1[A1]')

  def testTensorUnaryOp_ReturnsCorrectTypeAndShape(self):
    with utils.SaveCodeAsString() as code_saver:
      x1: Tensor0 = tf.zeros(())
      y1 = abs(x1)  # pylint: disable=unused-variable
      y2 = -x1  # pylint: disable=unused-variable
      x2: Tensor1[A1] = tf.zeros((1,))
      y3 = abs(x2)  # pylint: disable=unused-variable
      y4 = -x2  # pylint: disable=unused-variable

    inferred = utils.pytype_infer_types(_PREAMBLE + code_saver.code)

    self.assertEqual('Tensor0', inferred.y1)
    self.assertEqual('Tensor0', inferred.y2)
    self.assertEqual('Tensor1[A1]', inferred.y3)
    self.assertEqual('Tensor1[A1]', inferred.y4)

  def testBinaryOpWithScalar_InferredMatchesActualShape(self):
    with utils.SaveCodeAsString() as code_saver:
      x: Tensor2[A1, A2] = tf.zeros((1, 2))
      y1 = x + 1.0
      y2 = x - 1.0
      y3 = x / 1.0
      y4 = x * 1.0

    inferred = utils.pytype_infer_shapes(_PREAMBLE + code_saver.code)

    self.assertEqual(y1.shape, inferred.y1)
    self.assertEqual(y2.shape, inferred.y2)
    self.assertEqual(y3.shape, inferred.y3)
    self.assertEqual(y4.shape, inferred.y4)

  def testBinaryOpWithBroadcast_InferredMatchesActualShape(self):
    with utils.SaveCodeAsString() as code_saver:
      a: Tensor2[A1, A2] = tf.zeros((1, 2))
      b: Tensor1[A2] = tf.zeros((2,))
      y1 = a + b
      y2 = a - b
      y3 = a / b
      y4 = a * b

    inferred = utils.pytype_infer_shapes(_PREAMBLE + code_saver.code)

    self.assertEqual(y1.shape, inferred.y1)
    self.assertEqual(y2.shape, inferred.y2)
    self.assertEqual(y3.shape, inferred.y3)
    self.assertEqual(y4.shape, inferred.y4)

  def testBinaryOpWithSameShape_InferredMatchesActualShape(self):
    with utils.SaveCodeAsString() as code_saver:
      a: Tensor2[A1, A2] = tf.zeros((1, 2))
      b: Tensor2[A1, A2] = tf.zeros((1, 2))
      y1 = a + b
      y2 = a - b
      y3 = a / b
      y4 = a * b

    inferred = utils.pytype_infer_shapes(_PREAMBLE + code_saver.code)

    self.assertEqual(y1.shape, inferred.y1)
    self.assertEqual(y2.shape, inferred.y2)
    self.assertEqual(y3.shape, inferred.y3)
    self.assertEqual(y4.shape, inferred.y4)

  def testShapeAttribute_HasTypeTensorShape(self):
    with utils.SaveCodeAsString() as code_saver:
      x0 = tf.zeros(())
      x1 = tf.zeros((1,))
      x2 = tf.zeros((1, 2))
      x3 = tf.zeros((1, 2, 3))
      x4 = tf.zeros((1, 2, 3, 4))
      x5 = tf.zeros((1, 2, 3, 4, 5))
      x0_shape = x0.shape   # pylint: disable=unused-variable
      x1_shape = x1.shape   # pylint: disable=unused-variable
      x2_shape = x2.shape   # pylint: disable=unused-variable
      x3_shape = x3.shape   # pylint: disable=unused-variable
      x4_shape = x4.shape   # pylint: disable=unused-variable
      x5_shape = x5.shape   # pylint: disable=unused-variable

    inferred = utils.pytype_infer_types(_PREAMBLE + code_saver.code)

    self.assertEqual('tensorflow.TensorShape', inferred.x0_shape)
    self.assertEqual('tensorflow.TensorShape', inferred.x1_shape)
    self.assertEqual('tensorflow.TensorShape', inferred.x2_shape)
    self.assertEqual('tensorflow.TensorShape', inferred.x3_shape)
    self.assertEqual('tensorflow.TensorShape', inferred.x4_shape)
    self.assertEqual('tensorflow.TensorShape', inferred.x5_shape)

  def testShapeAttribute_HasLen(self):
    with utils.SaveCodeAsString() as code_saver:
      x = tf.zeros((1,))
      rank = len(x.shape)  # pylint: disable=unused-variable

    inferred = utils.pytype_infer_types(_PREAMBLE + code_saver.code)
    self.assertEqual('int', inferred.rank)


if __name__ == '__main__':
  unittest.main()



"""Tests for experimental/decorators.py."""

from typing import Dict, List, NamedTuple, Tuple
import unittest

from tensor_annotations.axes import Height
from tensor_annotations.axes import Width
from tensor_annotations.experimental import decorators
from tensor_annotations.tensorflow import Tensor1
from tensor_annotations.tensorflow import Tensor2
import tensorflow as tf


class DecoratorsArgsTests(unittest.TestCase):

  # ===== Tests for correct args. =====

  def test_correct_args_does_not_raise_exception(self):
    @decorators.verify_runtime_ranks_of_args_and_return
    def foo(
        x: Tensor1[Height],  # pylint: disable=unused-argument
        y: Tensor2[Height, Width],  # pylint: disable=unused-argument
    ):
      pass
    foo(tf.zeros([3]), tf.zeros([3, 5]))
    foo(x=tf.zeros([3]), y=tf.zeros([3, 5]))
    foo(tf.zeros([3]), y=tf.zeros([3, 5]))

  def test_correct_default_arg_does_not_raise_exception(self):
    @decorators.verify_runtime_ranks_of_args_and_return
    def foo(
        x: Tensor2[Height, Width] = tf.zeros([3, 5]),  # pylint: disable=unused-argument
    ):
      pass
    foo()
    foo(tf.zeros([3, 5]))

  # ===== Tests for unannotated args. =====

  def test_unannotated_arg_does_not_raise_exception(self):
    @decorators.verify_runtime_ranks_of_args_and_return
    def foo(x):  # pylint: disable=unused-argument
      pass
    foo(tf.zeros([3]))
    foo(x=tf.zeros([3]))

  def test_unannotated_and_correct_annotated_arg_does_not_raise_exception(self):
    @decorators.verify_runtime_ranks_of_args_and_return
    def foo(x, y: Tensor2[Height, Width]):  # pylint: disable=unused-argument
      pass
    foo(tf.zeros([3]), tf.zeros([3, 5]))
    foo(x=tf.zeros([3]), y=tf.zeros([3, 5]))
    foo(tf.zeros([3]), y=tf.zeros([3, 5]))

  def test_unrelated_annotation_does_not_raise_exception(self):
    @decorators.verify_runtime_ranks_of_args_and_return
    def foo(x: int):  # pylint: disable=unused-argument
      pass
    foo(0)

  def test_unrelated_generic_annotation_does_not_raise_exception(self):
    @decorators.verify_runtime_ranks_of_args_and_return
    def foo(x: List[int]):  # pylint: disable=unused-argument
      pass
    foo([0])

  # ===== Tests for incorrect args. =====

  def test_incorrect_arg_raises_exception(self):
    @decorators.verify_runtime_ranks_of_args_and_return
    def foo(x: Tensor2[Height, Width]):  # pylint: disable=unused-argument
      pass
    with self.assertRaises(TypeError):
      foo(tf.zeros([3]))
    with self.assertRaises(TypeError):
      foo(x=tf.zeros([3]))

  def test_incorrect_second_arg_raises_exception(self):
    @decorators.verify_runtime_ranks_of_args_and_return
    def foo(x, y: Tensor2[Height, Width]):  # pylint: disable=unused-argument
      pass
    with self.assertRaises(TypeError):
      foo(None, tf.zeros([3]))
    with self.assertRaises(TypeError):
      foo(x=None, y=tf.zeros([3]))
    with self.assertRaises(TypeError):
      foo(None, y=tf.zeros([3]))

  def test_incorrect_default_arg_raises_exception(self):
    @decorators.verify_runtime_ranks_of_args_and_return
    def foo(
        x: Tensor2[Height, Width] = tf.zeros([3]),  # pylint: disable=unused-argument
    ):
      pass
    with self.assertRaises(TypeError):
      foo()

  def test_incorrect_overridden_default_arg_raises_exception(self):
    @decorators.verify_runtime_ranks_of_args_and_return
    def foo(
        x: Tensor2[Height, Width] = tf.zeros([3, 5]),  # pylint: disable=unused-argument
    ):
      pass
    with self.assertRaises(TypeError):
      foo(tf.zeros([3]))
    with self.assertRaises(TypeError):
      foo(x=tf.zeros([3]))

  def test_non_tensor_arg_raises_exception(self):
    @decorators.verify_runtime_ranks_of_args_and_return
    def foo(x: Tensor2[Height, Width]):  # pylint: disable=unused-argument
      pass
    with self.assertRaises(TypeError):
      foo(None)

  # ==== Make sure 'normal' errors still behave as expected. =====

  def test_no_args_raises_exception(self):
    @decorators.verify_runtime_ranks_of_args_and_return
    def foo(x: Tensor2[Height, Width]):  # pylint: disable=unused-argument
      pass
    with self.assertRaises(TypeError):
      foo()  # pylint: disable=no-value-for-parameter

  def test_too_many_args_raises_exception(self):
    @decorators.verify_runtime_ranks_of_args_and_return
    def foo():
      pass
    with self.assertRaises(TypeError):
      foo(None)  # pylint: disable=too-many-function-args

  def test_missing_keyword_arg_raises_exception(self):
    @decorators.verify_runtime_ranks_of_args_and_return
    def foo():
      pass
    with self.assertRaises(TypeError):
      foo(x=None)  # pylint: disable=unexpected-keyword-arg


class DecoratorsReturnTests(unittest.TestCase):

  # ===== Cases which should *not* raise an exception. =====

  def test_return_value_is_preserved(self):
    @decorators.verify_runtime_ranks_of_args_and_return
    def foo():
      return 42
    self.assertEqual(foo(), 42)

  def test_correct_return_does_not_raise_exception(self):
    @decorators.verify_runtime_ranks_of_args_and_return
    def foo() -> Tensor2[Height, Width]:
      return tf.zeros([3, 5])
    foo()

  def test_non_tensor_annotation_does_not_raise_exception(self):
    @decorators.verify_runtime_ranks_of_args_and_return
    def foo() -> None:
      # This is incorrect, but it's not our job to catch it.
      return tf.zeros([3])
    foo()

  # ===== Cases which *should* raise an exception. =====

  def test_incorrect_return_raises_exception(self):
    @decorators.verify_runtime_ranks_of_args_and_return
    def foo() -> Tensor2[Height, Width]:
      return tf.zeros([3])
    with self.assertRaises(TypeError):
      foo()

  def test_non_tensor_return_value_raises_exception(self):
    @decorators.verify_runtime_ranks_of_args_and_return
    def foo() -> Tensor2[Height, Width]:
      return None
    with self.assertRaises(TypeError):
      foo()


class DecoratorsTreeTests(unittest.TestCase):

  def test_convert_tuple_type_is_correct(self):
    tree_type = Tuple[int, str]
    type_tree = decorators._tree_type_to_type_tree(tree_type)
    self.assertEqual(type_tree, (int, str))

  def test_convert_nested_tuple_type_is_correct(self):
    tree_type = Tuple[int, Tuple[float, str]]
    type_tree = decorators._tree_type_to_type_tree(tree_type)
    self.assertEqual(type_tree, (int, (float, str)))

  def test_convert_double_nested_tuple_type_is_correct(self):
    tree_type = Tuple[int, Tuple[float, Tuple[str]]]
    type_tree = decorators._tree_type_to_type_tree(tree_type)
    self.assertEqual(type_tree, (int, (float, (str,))))

  def test_convert_empty_tuple_type_is_correct(self):
    tree_type = Tuple[()]
    type_tree = decorators._tree_type_to_type_tree(tree_type)
    self.assertEqual(type_tree, ())

  def test_convert_unparameterised_tuple_remains_unparameterised(self):
    tree_type = Tuple[str, Tuple]
    type_tree = decorators._tree_type_to_type_tree(tree_type)
    self.assertEqual(type_tree, (str, Tuple))

  def test_convert_dict_type_is_correct(self):
    tree_type = Dict[str, int]
    type_tree = decorators._tree_type_to_type_tree(tree_type)
    self.assertEqual(type_tree, {str: int})

  def test_convert_nested_dict_type_is_correct(self):
    tree_type = Dict[str, Dict[int, float]]
    type_tree = decorators._tree_type_to_type_tree(tree_type)
    self.assertEqual(type_tree, {str: {int: float}})

  def test_convert_double_nested_dict_type_is_correct(self):
    tree_type = Dict[str, Dict[bool, Dict[int, float]]]
    type_tree = decorators._tree_type_to_type_tree(tree_type)
    self.assertEqual(type_tree, {str: {bool: {int: float}}})

  def test_convert_unparameterised_dict_remains_unparameterised(self):
    tree_type = Tuple[Dict]
    type_tree = decorators._tree_type_to_type_tree(tree_type)
    self.assertEqual(type_tree, (Dict,))

  def test_convert_named_tuple_type_is_correct(self):
    class T(NamedTuple):
      a: Tuple[int]
      b: Dict[str, bool]
    type_tree = decorators._tree_type_to_type_tree(T)
    self.assertEqual(type_tree, T(a=(int,), b={str: bool}))

  def test_convert_nested_named_tuple_type_is_correct(self):
    class T1(NamedTuple):
      a: Tuple[int]
    class T2(NamedTuple):
      b: T1
    type_tree = decorators._tree_type_to_type_tree(T2)
    self.assertEqual(
        type_tree,
        T2(
            b=T1(
                a=(int,),
            ),
        ),
    )

  def test_convert_double_nested_named_tuple_type_is_correct(self):
    class T1(NamedTuple):
      a: Tuple[int]
    class T2(NamedTuple):
      b: T1
    class T3(NamedTuple):
      c: T2
    type_tree = decorators._tree_type_to_type_tree(T3)
    self.assertEqual(
        type_tree,
        T3(
            c=T2(
                b=T1(
                    a=(int,),
                )
            ),
        ),
    )

  def test_convert_mixed_type_is_correct(self):
    class Trajectory(NamedTuple):
      actions: Dict[str, Tuple[float]]
      observations: Dict[str, Tuple[float]]

    class TrajectoriesAndParameters(NamedTuple):
      trajectories: Tuple[Trajectory]
      parameters: Tuple[float]

    type_tree = decorators._tree_type_to_type_tree(TrajectoriesAndParameters)
    self.assertEqual(
        type_tree,
        TrajectoriesAndParameters(
            trajectories=(
                Trajectory(
                    actions={str: (float,)},
                    observations={str: (float,)},
                ),
            ),
            parameters=(float,),
        )
    )


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Install script."""

import glob
import os
import pathlib
import setuptools

stub_files = (glob.glob('tensor_annotations/*.pyi') +
              glob.glob('tensor_annotations/library_stubs/**/*.pyi',
                        recursive=True))
# package_data expects paths to be relative to the package directory, so strip
# 'tensor_annotations/' from start of paths
stub_files = [
    os.path.join(*pathlib.Path(path).parts[1:])
    for path in stub_files
]

with open('README.md', 'r') as f:
    long_description = f.read()

setuptools.setup(
    name='tensor_annotations',
    version='1.0.0',
    description=('Enables annotations of tensor shapes in numerical computing '
                 'libraries. Includes type stubs for TensorFlow and JAX '
                 'describing how library functions change shapes.'),
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/deepmind/tensor_annotations',
    # Copybara takes care of moving files to 'tensor_annotations/'
    packages=[
        'tensor_annotations',
        'tensor_annotations/experimental',
        'tensor_annotations/tests',
    ],
    package_data={'tensor_annotations': stub_files + ['py.typed']},
    extras_require={'dev': [
        'absl-py',
        'pytype',
    ]})



# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Test helpers."""

import importlib
import inspect
import os
import pathlib
import re
import subprocess
import tempfile
import textwrap
import types
from typing import List
from typing import Optional


def run_pytype(code: str, check: bool) -> subprocess.CompletedProcess:  # pylint: disable=g-doc-args
  """Runs pytype on the specified code.

  Raises:
    subprocess.CalledProcessError if check=True and pytype return is non-zero

  Returns:
    A subprocess.CompletedProcess instance containing stdout
  """

  with tempfile.TemporaryDirectory() as tmp_dir:
    code_filename = os.path.join(tmp_dir, 'test.py')
    with open(code_filename, 'w') as f:
      f.write(code)

    pytype_path = pathlib.Path('pytype-single')
    tensor_annotations_dir = pathlib.Path(__file__).parent.parent
    stubs_dir = pathlib.Path(tmp_dir)
    _link_stubs(tensor_annotations_dir, stubs_dir)
    _generate_tensor_annotations_stubs(pytype_path, tensor_annotations_dir,
                                       stubs_dir)

    return subprocess.run(
        [str(pytype_path), '--pythonpath',
         str(stubs_dir), code_filename],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        check=check,  # Raise error if non-zero return code.
    )


def _link_stubs(tensor_annotations_dir: pathlib.Path, stubs_dir: pathlib.Path):
  """Link JAX/TensorFlow stubs to a place where pytype can find them."""
  google_internal = False
  if not google_internal:
    jax_module = importlib.import_module('jax-stubs')
    jax_stubs_dir = pathlib.Path(jax_module.__path__[0])
    tf_module = importlib.import_module('tensorflow-stubs')
    tf_stubs_dir = pathlib.Path(tf_module.__path__[0])

  for source, target in [
      # Library functions, e.g. tf.reduce_sum.
      (jax_stubs_dir, stubs_dir / 'jax'),
      (tf_stubs_dir, stubs_dir / 'tensorflow'),
      # Tensor functions, e.g. Tensor.__add__.
      (tensor_annotations_dir / 'jax.pyi',
       stubs_dir / 'tensor_annotations' / 'jax.pyi'),
      (tensor_annotations_dir / 'tensorflow.pyi',
       stubs_dir / 'tensor_annotations' / 'tensorflow.pyi')
  ]:
    if not os.path.exists(source):
      raise Exception(f"Stub file '{source}' does not exist")
    target.parent.mkdir(parents=True, exist_ok=True)
    target.symlink_to(source)


def _generate_tensor_annotations_stubs(pytype_path: pathlib.Path,
                                       tensor_annotations_dir: pathlib.Path,
                                       stubs_dir: pathlib.Path):
  """Generates stubs for tensor_annotations modules."""
  path = tensor_annotations_dir / 'axes.py'
  pyi_path = stubs_dir / 'tensor_annotations' / 'axes.pyi'
  pyi_path.parent.mkdir(parents=True, exist_ok=True)
  subprocess.run(
      [
          str(pytype_path),
          '-o',
          str(pyi_path),
          str(path),
      ],
      check=True,  # Raise error if non-zero return code.
  )


def assert_pytype_succeeds(code: str):
  try:
    process = run_pytype(code, check=True)
  except subprocess.CalledProcessError as e:
    print(e.stdout.decode())
    print(e.stderr.decode())
    raise e


def assert_pytype_fails(code: str):
  try:
    run_pytype(code, check=True)
  except subprocess.CalledProcessError:
    pass
  else:
    assert False, 'pytype did not raise error'


def pytype_infer_types(code: str) -> types.SimpleNamespace:
  """Runs pytype on `code`, returning inferred type of each variable.

  Args:
    code: Code to run pytype on. Should include assignment to one or more
          variables, e.g. 'x = jnp.zeros((3, 1))'. We collect a list of
          variables assigned to, and use pytype to infer type at the end of the
          code.

  Returns:
    A SimpleNamespace whose attributes map from variable names to types.

  Raises:
    Exception: If types could not be parsed from pytype output.

  For example, if `code` is
    x = 'foo'
    x = 2
    y = 3.0
  we return a SimpleNamespace `a` such that
    a.x = int
    a.y = float
  """
  # This may contain duplicates, but that's fine.
  var_names = re.findall(r'^([^: ]*).*=', code, re.MULTILINE)
  for var in var_names:
    code += f'\nreveal_type({var})'

  process = run_pytype(code, check=False)
  # If the test fails, make sure we can see why
  print(process.stdout.decode())
  print(process.stderr.decode())

  # We look at both stdout and stderr because pytype behaves differently
  # depending on whether we run the Google-internal version or the normal
  # version
  lines = (process.stdout.decode() + process.stderr.decode()).split('\n')

  return _parse_pytype_output(var_names, lines)


def _parse_pytype_output(var_names: List[str],
                         lines: List[str]) -> types.SimpleNamespace:
  """Parses the inferred type of each variable from pytype output."""
  reveal_type_lines = [l for l in lines if '[reveal-type]' in l]
  assert len(reveal_type_lines) == len(var_names)

  types_dict = {}
  for var, line in zip(var_names, reveal_type_lines):
    match = re.search(r'File "[^"]*", line \d+, in [^:]*: '
                      r'(.*) \[reveal-type\]', line)
    if match is None:
      raise Exception(f"Couldn't parse type from line: {line}")
    t = match.group(1)
    # Simplifies e.g. `tensor_annotations.jax.Array0` to just `Array0`
    t = re.sub(r'tensor_annotations.[^.]*\.', '', t)
    types_dict[var] = t

  return types.SimpleNamespace(**types_dict)


def _parse_mypy_output(var_names: List[str],
                       lines: List[str]) -> types.SimpleNamespace:
  """Parses the inferred type of each variable from Mypy output."""
  reveal_type_lines = [l for l in lines if 'Revealed type is' in l]
  assert len(reveal_type_lines) == len(var_names)

  types_dict = {}
  for var, line in zip(var_names, reveal_type_lines):
    match = re.search("Revealed type is '(.*)'", line)
    if match is None:
      raise Exception(f"Couldn't parse type from line: {line}")
    t = match.group(1)
    # Simplifies e.g. `tensor_annotations.jax.Array0` to just `Array0`
    t = re.sub(r'tensor_annotations.[^.]*\.', '', t)
    # Remove the '*' that Mypy suffixes types with if the types were inferred
    # using type variable substitution.
    t = t.replace('*', '')
    # Mypy will format axis types as e.g. `test.A1`. Get rid of the `test.`.
    t = re.sub(r'test.(A\d+)', r'\1', t)
    # Mypy will format unparameterised generics as e.g. `Tensor1[Any]`, but
    # we wrote tests assuming they'd be formatted as just `Tensor1`, so get
    # rid of the `[Any]`.
    t = t.replace('[Any]', '')
    types_dict[var] = t

  return types.SimpleNamespace(**types_dict)


def pytype_infer_shapes(code: str) -> types.SimpleNamespace:
  # pylint: disable=g-doc-args,g-doc-return-or-yield,g-doc-exception
  """Runs pytype on `code`, returning inferred shape of each variable.

  Note that shapes are inferred based on the axis labels: axis label 'A1' is
  assumed to represent a dimension of size 1, 'A2' a dimension of size 2, and so
  on. For example, we assume that a tensor of type Tensor2[A1, A2] has shape
  (1, 2).

  For example, if `code` is
    x: tf.Tensor2[A3, A5] = tf.zeros((3, 5))
    y = tf.transpose(x)  # tf.Tensor2[A5, A3]
  we return a SimpleNamespace `a` such that
    a.x = (3, 5)
    a.y = (5, 3)

  This helper function exists so that we can easily compare the shape of the
  real outputs of shape-changing functions (e.g. tf.transpose) to the shape
  inferred by the type checker using our stubs, to confirm that our stubs
  are correct.

  See `pytype_infer_types` for more info.
  """
  types_namespace = pytype_infer_types(code)

  shapes_dict = {}
  var_names = [d for d in dir(types_namespace)
               if not d.startswith('_')]
  for var in var_names:
    t = getattr(types_namespace, var)
    if t.endswith('Array0') or t.endswith('Tensor0'):
      shape = ()
    elif t == 'Any':
      shape = 'Any'
    else:
      match = re.search(r'\[(.*)\]', t)
      if match is None:
        # Could have been e.g. a float.
        continue
      axis_types = match.group(1)  # e.g. 'A1, A2'
      shape_str_list = axis_types.replace('A', '').split(', ')
      shape = tuple(int(s) for s in shape_str_list)
    shapes_dict[var] = shape

  return types.SimpleNamespace(**shapes_dict)


class SaveCodeAsString:
  r"""Saves code executed within the context manager.

  Indentation is automatically adjusted such that the first line has no
  indenation in the saved code. For example, if used as follows:
    with SaveCodeString() as code_saver:
      foo = 'bar'
      f()
  then `code_saver.code` would contain "foo = 'bar'\nf()".
  """

  def __init__(self):
    self._frame_where_entered = None
    self._frame_where_exited = None
    self.code: Optional[str] = None

  def __enter__(self):
    self._frame_where_entered = inspect.stack()[1]
    return self

  def __exit__(self, *_):
    self._frame_where_exited = inspect.stack()[1]
    with open(self._frame_where_entered.filename, 'r') as f:
      lines = f.readlines()
    start = self._frame_where_entered.lineno
    end = self._frame_where_exited.lineno
    lines = lines[start:end]
    self.code = ''.join(lines).rstrip()
    self.code = textwrap.dedent(self.code)
